"""Fileman gRPC services."""
import logging
import os
import random
import string
import sys
from pathlib import Path
from uuid import UUID

from fred_api.fileman.service_fileman_grpc_pb2 import ReadReply, ReadRequest, StatReply, StatRequest
from fred_api.fileman.service_fileman_grpc_pb2_grpc import FileServicer as BaseFileServicer
from google.protobuf.message import Message
from google.protobuf.text_format import MessageToString
from grpc import StatusCode
from grpc._server import _Context
from sqlalchemy.exc import DBAPIError
from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from .models import File

_LOGGER = logging.getLogger(__name__)
MESSAGE_MAX_LENGTH = 256


def _make_call_id() -> str:
    """Return unique call identifier to match subsequent logs."""
    alphabet = string.ascii_letters + string.digits
    return ''.join(random.choice(alphabet) for _ in range(4))  # nosec


def message_formatter(message: Message) -> str:
    """Format a gRPC Message to string."""
    out = MessageToString(message, as_one_line=True)
    if len(out) > MESSAGE_MAX_LENGTH:
        return out[:MESSAGE_MAX_LENGTH - 3] + '...'
    else:
        return out


class FileServicer(BaseFileServicer):
    """File gRPC service.

    Attributes:
        storage: A path to the root of the storage.
        session_cls: SQLAlchemy Session class to be used for queries.
    """

    def __init__(self, storage: str, session_cls: Session = None):
        self.storage = Path(storage)
        self.session_cls = session_cls or sessionmaker()

    def stat(self, request: StatRequest, context: _Context) -> StatReply:
        """Return file metadata."""
        call_id = _make_call_id()
        _LOGGER.debug("[%s] FileServicer.stat(%s)", call_id, message_formatter(request))
        try:
            uuid = UUID(request.uuid.value)
        except ValueError as error:
            _LOGGER.debug("[%s] FileServicer.stat failed with INVALID_ARGUMENT: %r", call_id, error)
            context.abort(StatusCode.INVALID_ARGUMENT, "'{}' is not a valid UUID".format(request.uuid.value))

        session = self.session_cls()
        try:
            file = session.query(File).filter_by(uuid=uuid).one()
        except NoResultFound as error:
            _LOGGER.debug("[%s] FileServicer.stat failed with NOT_FOUND: %r", call_id, error)
            context.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))
        except DBAPIError as error:
            _LOGGER.error("[%s] FileServicer.stat failed with %r", call_id, error)
            context.abort(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error))
        finally:
            session.close()
        reply = StatReply()
        reply.data.create_datetime.FromDatetime(file.create_datetime)
        reply.data.size = file.size
        reply.data.mimetype = file.mimetype
        reply.data.name = file.name
        _LOGGER.debug("[%s] FileServicer.stat returned %s", call_id, message_formatter(reply))
        return reply

    def _get_chunk_size(self, size: int, file_size: int) -> int:
        """Return a reasonable chunk size.

        Arguments:
            size: Requested chunk size
            file_size: Actual size of a requested file
        """
        if not size:
            # Size not set, return everything.
            return -1
        else:
            # Limit requested size to the actual file size. Don't go over sys.maxsize, may cause as overflow error.
            return min(size, file_size, sys.maxsize)

    def read(self, request: ReadRequest, context: _Context) -> ReadReply:
        """Return file content."""
        call_id = _make_call_id()
        _LOGGER.debug("[%s] FileServicer.read(%s)", call_id, message_formatter(request))
        try:
            uuid = UUID(request.uuid.value)
        except ValueError as error:
            _LOGGER.debug("[%s] FileServicer.read failed with INVALID_ARGUMENT: %r", call_id, error)
            context.abort(StatusCode.INVALID_ARGUMENT, "'{}' is not a valid UUID".format(request.uuid.value))

        session = self.session_cls()
        try:
            file_path = session.query(File.path).filter_by(uuid=uuid).one()
        except NoResultFound as error:
            _LOGGER.debug("[%s] FileServicer.read failed with NOT_FOUND: %r", call_id, error)
            context.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))
        except DBAPIError as error:
            _LOGGER.error("[%s] FileServicer.read failed with %r", call_id, error)
            context.abort(StatusCode.FAILED_PRECONDITION, "Database error: {}".format(error))
        finally:
            session.close()

        try:
            path = self.storage.joinpath(file_path[0])
            # XXX: Explicitly convert Path to str to support Python 3.5. `getsize` accepts Path since 3.6.
            file_size = os.path.getsize(str(path))
            chunk_size = self._get_chunk_size(request.size, file_size)
            # XXX: Explicitly convert Path to str to support Python 3.5. `open` accepts Path since 3.6.
            with open(str(path), mode='rb') as reader:
                while True:
                    reply = ReadReply()
                    chunk = reader.read(chunk_size)
                    reply.data.data = chunk
                    _LOGGER.debug("[%s] FileServicer.read returned %s", call_id, message_formatter(reply))
                    yield reply

                    if chunk_size == -1 or chunk_size == file_size or len(chunk) < chunk_size:
                        # If the file is read, terminate the generator.
                        _LOGGER.debug("[%s] FileServicer.read finished", call_id)
                        return
        except OSError as error:
            _LOGGER.error("[%s] FileServicer.read failed with FAILED_PRECONDITION: %r", call_id, error)
            context.abort(StatusCode.FAILED_PRECONDITION, "File {} is not available: {}".format(uuid, error))
