import os
import sys
from datetime import datetime
from tempfile import TemporaryDirectory
from typing import Iterable
from unittest import TestCase
from unittest.mock import Mock, call, sentinel
from uuid import UUID

from fred_api.fileman.service_fileman_grpc_pb2 import ReadReply, ReadRequest, StatReply, StatRequest
from grpc import StatusCode
from pytz import UTC
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from testfixtures import LogCapture, StringComparison

from fileman.models import BaseModel, File
from fileman.service import MESSAGE_MAX_LENGTH, FileServicer, message_formatter


class MessageFormatterTest(TestCase):
    def test_short(self):
        message = ReadRequest()
        message.uuid.value = 'Arnold Rimmer'
        self.assertEqual(message_formatter(message), 'uuid { value: "Arnold Rimmer" }')

    def test_long(self):
        message = ReadReply()
        message.data.data = b"I am fish! " * 400

        out = message_formatter(message)

        self.assertEqual(len(out), MESSAGE_MAX_LENGTH)
        self.assertEqual(out, StringComparison(r'data { data: "I am fish! .*\.\.\.'))


class FileServicerTest(TestCase):
    def setUp(self):
        # We mess the database structure in some tests. Create database every time.
        self.engine = create_engine('sqlite:///:memory:')
        BaseModel.metadata.create_all(self.engine)  # type: ignore
        self.connection = self.engine.connect()
        self.session_cls = sessionmaker(bind=self.connection)
        self.session = self.session_cls()

        self.tmp_dir = TemporaryDirectory()
        self.service = FileServicer(self.tmp_dir.name, self.session_cls)
        self.log_handler = LogCapture('fileman', propagate=False)

    def tearDown(self):
        self.log_handler.uninstall()
        self.tmp_dir.cleanup()

    def test_stat_empty(self):
        request = StatRequest()
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            self.service.stat(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls, [call.abort(StatusCode.INVALID_ARGUMENT, "'' is not a valid UUID")])

    def test_stat_invalid_uuid(self):
        request = StatRequest()
        request.uuid.value = 'invalid-uuid'
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            self.service.stat(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls,
                         [call.abort(StatusCode.INVALID_ARGUMENT, "'invalid-uuid' is not a valid UUID")])

    def test_stat_db_error(self):
        # Test DB error occurs on database level - dropped table
        BaseModel.metadata.drop_all(bind=self.connection)  # type: ignore
        request = StatRequest()
        request.uuid.value = str(UUID(int=42))
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            self.service.stat(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls,
                         [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("Database error: .*"))])

    def test_stat_unknown(self):
        uuid = UUID(int=42)
        request = StatRequest()
        request.uuid.value = str(uuid)
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            self.service.stat(request, context)

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls, [call.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))])

    def test_stat(self):
        uuid = UUID(int=42)
        name = 'example.txt'
        create_datetime = datetime(2019, 12, 11, 10, 9, 8, tzinfo=UTC)
        mimetype = 'text/plain'
        size = 1024
        afile = File(uuid=uuid, create_datetime=create_datetime, path='/', mimetype=mimetype, size=size, name=name)
        self.session.add(afile)
        self.session.commit()

        request = StatRequest()
        request.uuid.value = str(uuid)
        result = self.service.stat(request, sentinel.context)

        reply = StatReply()
        reply.data.create_datetime.FromDatetime(create_datetime)
        reply.data.size = size
        reply.data.mimetype = mimetype
        reply.data.name = name
        self.assertEqual(result, reply)

    def test_read_empty(self):
        request = ReadRequest()
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            tuple(self.service.read(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls, [call.abort(StatusCode.INVALID_ARGUMENT, "'' is not a valid UUID")])

    def test_read_invalid_uuid(self):
        request = ReadRequest()
        request.uuid.value = 'invalid-uuid'
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            tuple(self.service.read(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls,
                         [call.abort(StatusCode.INVALID_ARGUMENT, "'invalid-uuid' is not a valid UUID")])

    def test_read_db_error(self):
        # Test DB error occurs on database level - dropped table
        BaseModel.metadata.drop_all(bind=self.connection)  # type: ignore
        request = ReadRequest()
        request.uuid.value = str(UUID(int=42))
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            tuple(self.service.read(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls,
                         [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("Database error: .*"))])

    def test_read_no_file(self):
        # Test case, when the file is not found in the storage.
        uuid = UUID(int=42)
        path = 'example-path.txt'
        afile = File(uuid=uuid, path=path, size=1024, name='example.txt')
        self.session.add(afile)
        self.session.commit()

        request = ReadRequest()
        request.uuid.value = str(UUID(int=42))
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            tuple(self.service.read(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(
            context.mock_calls,
            [call.abort(StatusCode.FAILED_PRECONDITION, StringComparison("File .* is not available: .*"))])

    def test_read_unknown(self):
        uuid = UUID(int=42)
        request = ReadRequest()
        request.uuid.value = str(uuid)
        context = Mock(name='context')
        context.abort.side_effect = Exception(sentinel.abort)

        with self.assertRaises(Exception) as catcher:
            tuple(self.service.read(request, context))

        self.assertEqual(catcher.exception.args, (sentinel.abort, ))
        self.assertEqual(context.mock_calls, [call.abort(StatusCode.NOT_FOUND, "File {} not found".format(uuid))])

    def _test_read(self, content: bytes, replies: Iterable[ReadReply], size: int = 0):
        uuid = UUID(int=42)
        path = 'example-path.txt'
        with open(os.path.join(self.tmp_dir.name, path), mode='wb') as tmp_file:
            tmp_file.write(content)
        afile = File(uuid=uuid, path=path, size=1024, name='example.txt')
        self.session.add(afile)
        self.session.commit()

        request = ReadRequest()
        request.uuid.value = str(uuid)
        request.size = size
        result = self.service.read(request, sentinel.context)

        self.assertEqual(tuple(result), replies)

    def test_read(self):
        content = b'Rimmer is a smeg head!'
        reply = ReadReply()
        reply.data.data = content
        self._test_read(content, (reply, ))

    def test_read_chunked(self):
        content = b'Rimmer is a smeg head!'
        reply1 = ReadReply()
        reply1.data.data = content[:12]
        reply2 = ReadReply()
        reply2.data.data = content[12:]
        self._test_read(content, replies=(reply1, reply2), size=12)

    def test_read_empty_file(self):
        reply = ReadReply()
        reply.data.data = b''
        self._test_read(b'', (reply, ))

    def test_read_supersized(self):
        # Test unreasonably big size works as well.
        content = b'Rimmer is a smeg head!'
        reply = ReadReply()
        reply.data.data = content
        self._test_read(content, (reply, ), size=sys.maxsize + 1)
