import os
import tempfile
from contextlib import contextmanager
from unittest import TestCase
from unittest.mock import call, mock_open, patch, sentinel

import yaml
from easysettings import YAMLSettings
from testfixtures import LogCapture

from fileman.settings import DEFAULT_SETTINGS, Settings, SettingsProxy

from .utils import TestSettings


@contextmanager
def set_environ(**kwargs):
    """Clean the environment and set only provided variables."""
    with patch.dict(os.environ) as mock:
        os.environ.clear()
        os.environ.update(kwargs)
        yield mock


class SettingsTest(TestCase):
    home_call = call('/home/holly/.fred/fileman.conf', 'r')
    system_call = call('/etc/fred/fileman.conf', 'r')

    def test_getitem(self):
        settings = Settings()
        settings._settings = YAMLSettings({'secret': 'Gazpacho!'})
        self.assertEqual(settings['secret'], 'Gazpacho!')

    def test_getitem_unknown(self):
        settings = Settings()
        settings._settings = YAMLSettings()
        with self.assertRaisesRegex(KeyError, 'secret'):
            settings['secret']

    def test_setitem(self):
        settings = Settings()
        settings._settings = YAMLSettings({'secret': 'Gazpacho!'})
        settings['secret'] = 'Wine cellar'
        self.assertEqual(settings['secret'], 'Wine cellar')

    def test_setitem_unknown(self):
        settings = Settings()
        settings._settings = YAMLSettings()
        settings['secret'] = 'Wine cellar'
        self.assertEqual(settings['secret'], 'Wine cellar')

    def test_items_empty(self):
        settings = Settings()
        settings._settings = YAMLSettings()
        self.assertEqual(settings.items(), {}.items())

    def test_items(self):
        settings = Settings()
        settings._settings = YAMLSettings({'secret': 'Gazpacho!'})
        self.assertEqual(settings.items(), {'secret': 'Gazpacho!'}.items())

    def assertDefaults(self, settings):
        """Check settings contain all default settings."""
        for key, value in DEFAULT_SETTINGS.items():
            self.assertEqual(settings[key], value)

    def test_load_environ(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'secret': 'Gazpacho!'}, config_io)

        settings = Settings()
        with set_environ(FILEMAN_CONFIG=config_file.name):
            settings.load()

        self.assertEqual(settings['secret'], 'Gazpacho!')
        self.assertDefaults(settings)

    def test_load_environ_unknown(self):
        settings = Settings()
        with set_environ(FILEMAN_CONFIG='/does/not/exist'):
            with self.assertRaisesRegex(FileNotFoundError, '/does/not/exist'):
                settings.load()

    def test_load_argv(self):
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'secret': 'Wine cellar'}, config_io)

        settings = Settings()
        with set_environ():
            settings.load(config_file.name)

        self.assertEqual(settings['secret'], 'Wine cellar')
        self.assertDefaults(settings)

    def test_load_argv_unknown(self):
        settings = Settings()
        with set_environ():
            with self.assertRaisesRegex(FileNotFoundError, '/does/not/exist'):
                settings.load('/does/not/exist')

    def test_load_home(self):
        data = yaml.dump({'secret': 'Rusty gate'})

        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                settings.load()

        self.assertEqual(settings['secret'], 'Rusty gate')
        self.assertDefaults(settings)
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertNotIn(self.system_call, open_mock.mock_calls)

    def test_load_system(self):
        data = yaml.dump({'secret': 'Androids'})

        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open',
                       side_effect=[FileNotFoundError, mock_open(read_data=data).return_value]) as open_mock:
                settings.load()

        self.assertEqual(settings['secret'], 'Androids')
        self.assertDefaults(settings)
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertIn(self.system_call, open_mock.mock_calls)

    def test_load_no_settings(self):
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open', side_effect=FileNotFoundError) as open_mock:
                settings.load()

        self.assertDefaults(settings)
        self.assertIn(self.home_call, open_mock.mock_calls)
        self.assertIn(self.system_call, open_mock.mock_calls)

    def test_load_no_reload(self):
        # Test reload is not allowed.
        config_file = tempfile.NamedTemporaryFile()
        with open(config_file.name, 'w') as config_io:
            yaml.dump({'secret': 'Gazpacho!'}, config_io)

        settings = Settings()
        with set_environ(FILEMAN_CONFIG=config_file.name):
            settings.load()
            with self.assertRaisesRegex(RuntimeError, 'Settings already loaded'):
                settings.load()

    def test_getitem_load(self):
        # Test settings loads if not already loaded.
        data = yaml.dump({'secret': 'Gazpacho!'})
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                self.assertEqual(settings['secret'], 'Gazpacho!')

        self.assertIn(self.home_call, open_mock.mock_calls)

    def test_setitem_load(self):
        # Test settings loads if not already loaded.
        data = yaml.dump({'secret': 'Gazpacho!'})
        settings = Settings()
        with set_environ(HOME='/home/holly'):
            with patch('builtins.open', mock_open(read_data=data)) as open_mock:
                settings['secret'] = 'Wine cellar'

        self.assertEqual(settings['secret'], 'Wine cellar')
        self.assertIn(self.home_call, open_mock.mock_calls)

    def test_set_options(self):
        settings = Settings()

        settings.set_options({'--port': sentinel.port})

        self.assertEqual(settings['grpc_port'], sentinel.port)

    def test_set_options_defaults(self):
        # Test default values from argv are ignored.
        settings = Settings()

        settings.set_options({'--port': None})

        self.assertEqual(settings['grpc_port'], 50051)

    def test_sentry(self):
        dsn = 'Gazpacho!'
        config_file = tempfile.NamedTemporaryFile(mode='w')
        yaml.dump({'sentry_dsn': dsn}, config_file)

        settings = Settings()
        with patch('fileman.settings.sentry_sdk') as sentry_mock:
            with set_environ():
                settings.load(config_file.name)

        self.assertEqual(sentry_mock.mock_calls, [call.init(dsn=dsn)])

    def test_sentry_no_sdk(self):
        config_file = tempfile.NamedTemporaryFile(mode='w')
        yaml.dump({'sentry_dsn': 'Gazpacho!'}, config_file)

        settings = Settings()
        with patch('fileman.settings.sentry_sdk', new=None):
            with set_environ():
                with LogCapture('fileman', propagate=False) as log:
                    settings.load(config_file.name)

        log.check(('fileman.settings', 'WARNING', 'Sentry set up, but sentry-sdk not available.'))


class SettingsProxyTest(TestCase):
    def test_getitem(self):
        settings = TestSettings()
        settings['secret'] = 'Gazpacho!'
        proxy = SettingsProxy(settings)
        self.assertEqual(proxy['secret'], 'Gazpacho!')

    def test_setitem(self):
        settings = TestSettings()
        proxy = SettingsProxy(settings)
        proxy['secret'] = 'Gazpacho!'
        self.assertEqual(settings['secret'], 'Gazpacho!')

    def test_getattr(self):
        settings = TestSettings()
        settings['secret'] = 'Gazpacho!'
        proxy = SettingsProxy(settings)
        self.assertEqual(proxy.items(), settings.items())
